#include <stdio.h>
#include "All.h"
#include "MACLib.h"
#include <Carbon/Carbon.h>
#include <CoreAudio/AudioHardware.h>
#include <AudioToolbox/AudioConverter.h>
#include <unistd.h>

//#define SIZEOF_BUFFER 131072

static OSStatus
macosx_audio_out_callback (AudioDeviceID device, const AudioTimeStamp* current_time,
	const AudioBufferList* data_in, const AudioTimeStamp* time_in,
	AudioBufferList*	data_out, const AudioTimeStamp* time_out,
	void* client_data);

struct soundInfo
{
	AudioStreamBasicDescription outFormat;
	AudioStreamBasicDescription inFormat;
	AudioDeviceID device;
	AudioConverterRef converter;
	int done_playing;
	IAPEDecompress *ape_data;
	unsigned int nbytes;
	UInt32 			buf_size;
	int errnum;
	char *buffer;
	float *buffer2;
	int bufferIsFull, buffer2IsEmpty;
	int lastPlay;
	UInt32 bufferRest, myBufferRest;
	UInt32 nextReadSize;
	float *nextStart;
	int lastBuffer;
	int quiet;
	int softBufSize;
	int softBufSize2;
	unsigned int percent;
	
	short channels;
	short bytesPerChannel;
	short bitsPerChannel;
	int samplingRate;
	unsigned int approxFileSize;
	unsigned int dataBytes;
};

int main(int argc, char *argv[])
{
	FILE *fp;
	struct soundInfo sinfo;
	OSStatus	err ;
	UInt32		count;
	char	*buffer_ptr, *buffer_tmp, opt;
	int i=0;
	int blockper, received;
	extern char *optarg ;
	extern int optind;
	
	sinfo.quiet = 0;
	sinfo.buf_size = 32768;
	
	while((opt = getopt(argc,argv,"qb:")) != -1) {
		switch(opt) {
		  case 'q' :
			sinfo.quiet = 1;
			break;
		  case 'b' :
			sinfo.buf_size = atoi(optarg);
			break;
		}
		i++;
	}
	
	if(argc <= i+1) {
		printf("apeplay by tmkk\n usage: %s <ape file>\n",argv[0]);
		return 0;
	}
	
	if((fp=fopen(argv[optind],"rb")) == NULL) {
		fprintf(stderr,"cannot open file.\n");
		return 0;
	}
	fclose(fp);
	
	sinfo.device = kAudioDeviceUnknown;
	
	/*  get the default output device for the HAL */
	count = sizeof (AudioDeviceID) ;
	if ((err = AudioHardwareGetProperty (kAudioHardwarePropertyDefaultOutputDevice,
				&count, (void *) &(sinfo.device))) != noErr)
	{	printf ("AudioHardwareGetProperty failed.\n") ;
		return 0;
	} ;
	
	IAPEDecompress *decomp = CreateIAPEDecompress(argv[optind], &(sinfo.errnum));
	if(sinfo.errnum) {
		fprintf(stderr, "file format is not APE.\n");
		return 0;
	}
	
	sinfo.channels = decomp->GetInfo(APE_INFO_CHANNELS,0,0);
	sinfo.bitsPerChannel = decomp->GetInfo(APE_INFO_BITS_PER_SAMPLE,0,0);
	sinfo.bytesPerChannel = decomp->GetInfo(APE_INFO_BYTES_PER_SAMPLE,0,0);
	sinfo.samplingRate = decomp->GetInfo(APE_INFO_SAMPLE_RATE,0,0);
	sinfo.dataBytes = decomp->GetInfo(APE_INFO_WAV_DATA_BYTES,0,0);
	
	
	sinfo.ape_data = decomp;
	sinfo.nbytes = 0;
	
	/*  get a description of the data format used by the default device */
	
	count = sizeof (AudioStreamBasicDescription) ;
	if ((err = AudioDeviceGetProperty (sinfo.device, 0, false, kAudioDevicePropertyStreamFormat,
				&count, &(sinfo.outFormat))) != noErr)
	{	printf ("AudioDeviceGetProperty (kAudioDevicePropertyStreamFormat) failed.\n") ;
		return 0;
	} ;
	
	//printf("%d,%d,%d,%d,%d\n",sinfo.format.mFormatFlags,sinfo.format.mBytesPerPacket,sinfo.format.mFramesPerPacket,sinfo.format.mBytesPerFrame,sinfo.format.mChannelsPerFrame);
	
	/*
	if(sinfo.bytesPerChannel != 2) {
		printf("unsupported bit depth.\n");
		delete decomp;
		return 0;
	}
	*/
	sinfo.inFormat = sinfo.outFormat;
	//sinfo.inFormat.mSampleRate = (double)sinfo.samplingRate;
	if(sinfo.bitsPerChannel == 8) 
		sinfo.inFormat.mFormatFlags=kLinearPCMFormatFlagIsPacked;
	else
		sinfo.inFormat.mFormatFlags=kLinearPCMFormatFlagIsSignedInteger|kLinearPCMFormatFlagIsPacked;
	sinfo.inFormat.mBytesPerPacket=sinfo.bytesPerChannel*sinfo.channels;
	sinfo.inFormat.mBytesPerFrame=sinfo.bytesPerChannel*sinfo.channels;
	sinfo.inFormat.mChannelsPerFrame=sinfo.channels;
	sinfo.inFormat.mBitsPerChannel=sinfo.bitsPerChannel;
	
	if ((err = AudioDeviceSetProperty (sinfo.device, NULL, 0, false, kAudioDevicePropertyStreamFormat,
					sizeof (AudioStreamBasicDescription), &(sinfo.outFormat))) != noErr)
	{	printf ("AudioDeviceSetProperty (kAudioDevicePropertyStreamFormat) failed.\n") ;
		return 0;
	} ;
	
	if(AudioConverterNew(&(sinfo.inFormat), &(sinfo.outFormat), &(sinfo.converter))) {
        printf("AudioConverterNew failed\n");
        return 0;
    }
	
	
	/*  get the buffersize that the default device uses for IO */
	count = sizeof (UInt32) ;
	if(sinfo.buf_size) {
		if ((err = AudioDeviceSetProperty(sinfo.device, 0, 0, false, kAudioDevicePropertyBufferSize,
			count, &(sinfo.buf_size))) != noErr) {
			printf ("hardware buffer size %d is not allowed.\n",sinfo.buf_size) ;
			sinfo.buf_size = 0;
		}
	}
	
	if(!sinfo.buf_size) {
		for(i=2048;i<=32768;i=i+1024) {
			sinfo.buf_size = i;
			if ((err = AudioDeviceSetProperty(sinfo.device, 0, 0, false, kAudioDevicePropertyBufferSize,
				count, &(sinfo.buf_size))) != noErr) {
				//printf ("hardware buffer size %d is not allowed.\n",sinfo.buf_size) ;
				sinfo.buf_size = i-1024;
				AudioDeviceSetProperty(sinfo.device, 0, 0, false, kAudioDevicePropertyBufferSize,count, &(sinfo.buf_size));
				break;
			}
		}
	}
	
	if ((err = AudioDeviceGetProperty (sinfo.device, 0, false, kAudioDevicePropertyBufferSize,
		&count, &(sinfo.buf_size))) != noErr)
	{	printf ("AudioDeviceGetProperty (AudioDeviceGetProperty) failed.\n") ;
		return 0;
	} ;
	
		/*  we want linear pcm */
	if (sinfo.outFormat.mFormatID != kAudioFormatLinearPCM)
		return 0;
	
	if ((err = AudioDeviceAddIOProc (sinfo.device, macosx_audio_out_callback, 
				(void *) &sinfo)) != noErr)
	{	printf ("AudioDeviceAddIOProc failed.\n") ;
			return 0;
	} ;
	
	printf("Playing %s...\n",argv[optind]);
	printf("Format: %dHz,%dbit,%dch\n",sinfo.samplingRate,sinfo.bitsPerChannel,sinfo.channels);
	float ver = decomp->GetInfo(APE_INFO_FILE_VERSION,0,0)/1000.0;
	switch(decomp->GetInfo(APE_INFO_COMPRESSION_LEVEL,0,0)) {
	  case 1000:
		printf("Compression level: fast (%1.2f)\n",ver);
		break;
	  case 2000:
		printf("Compression level: normal (%1.2f)\n",ver);
		break;
	  case 3000:
		printf("Compression level: high (%1.2f)\n",ver);
		break;
	  case 4000:
		printf("Compression level: extra high (%1.2f)\n",ver);
		break;
	  case 5000:
		printf("Compression level: insane (%1.2f)\n",ver);
		break;
	  default:
		printf("Compression level: unknown (%1.2f)\n",ver);
	}
	printf("hardware buffer size: %d\n",sinfo.buf_size);
	
	sinfo.softBufSize = sinfo.buf_size * 4;
	if(sinfo.softBufSize > 98304) sinfo.softBufSize = sinfo.buf_size * 3;
	
	sinfo.softBufSize2 = sinfo.softBufSize*8/sinfo.channels/sinfo.bytesPerChannel;
	//printf("%d\n",sinfo.softBufSize2);
	
	if((sinfo.buffer = (char *)malloc(sinfo.softBufSize)) == NULL) {
		fprintf(stderr,"malloc error\n");
		return 0;
	}
	if((sinfo.buffer2 = (float *)malloc(sinfo.softBufSize2)) == NULL) {
		fprintf(stderr,"malloc error\n");
		return 0;
	}
	if((buffer_tmp = (char *)malloc(sinfo.softBufSize)) == NULL) {
		fprintf(stderr,"malloc error\n");
		return 0;
	}
	buffer_ptr = (char *)sinfo.buffer;
	sinfo.done_playing = 1 ;
	blockper = sinfo.softBufSize / sinfo.bytesPerChannel / sinfo.channels;
	sinfo.buffer2IsEmpty = 1;
	sinfo.lastPlay = 0;
	sinfo.nextReadSize = sinfo.buf_size;
	sinfo.nextStart = sinfo.buffer2;
	sinfo.bufferIsFull = 1;
	sinfo.myBufferRest = sinfo.softBufSize;
	sinfo.percent = 101;
	//printf("%d\n",sinfo.nextReadSize);
	
	decomp->GetData(buffer_tmp, blockper, &received);
	if(sinfo.errnum) {
		fprintf(stderr, "decoding error.\n");
		free(sinfo.buffer);
		free(sinfo.buffer2);
		free(buffer_tmp);
		delete decomp;
		return 0;
	}
	
	while(1) {
		if(received<=0) {
			break;
		}
		
		memcpy(buffer_ptr, buffer_tmp, received*sinfo.bytesPerChannel*sinfo.channels);
		sinfo.bufferRest = received * sinfo.bytesPerChannel * sinfo.channels;
		sinfo.bufferIsFull = 1;
		
		if(received<blockper) {
			break;
		}
		
		decomp->GetData(buffer_tmp, blockper, &received);
		if(sinfo.errnum) {
			fprintf(stderr, "decoding error.\n");
			free(sinfo.buffer);
			free(sinfo.buffer2);
			free(buffer_tmp);
			delete decomp;
			return 0;
		}
		
		if(sinfo.done_playing == 1) {
			sinfo.done_playing = 0 ;
			err = AudioDeviceStart (sinfo.device, macosx_audio_out_callback) ;
			if	(err != noErr)
				return 0;
		}
		
		while(sinfo.bufferIsFull){ //queue full
			usleep(10000); //0.01sec
		}
		
	}
	
	while (sinfo.done_playing == 0)
		usleep(10000) ;
	
	if ((err = AudioDeviceStop (sinfo.device, macosx_audio_out_callback)) != noErr)
	{	printf ("AudioDeviceStop failed.\n") ;
		return 0;
	} ;
    
	err = AudioDeviceRemoveIOProc (sinfo.device, macosx_audio_out_callback) ;
	if (err != noErr)
	{	printf ("AudioDeviceRemoveIOProc failed.\n") ;
		return 0;
	} ;
	
	free(buffer_tmp);
	free(sinfo.buffer);
	free(sinfo.buffer2);
	delete decomp;
	
	return 0;
}

static OSStatus
macosx_audio_out_callback (AudioDeviceID device, const AudioTimeStamp* current_time,
	const AudioBufferList* data_in, const AudioTimeStamp* time_in,
	AudioBufferList*	data_out, const AudioTimeStamp* time_out,
	void* client_data)
{	soundInfo	*audio_data ;
	unsigned int i,percent;
	float *buffer; /*, scale = (1.0f / SHRT_MAX);*/
	//short *buffer_shrt;
	UInt32 converted;
	OSStatus err;
	
	audio_data = (soundInfo *)client_data;
	
	converted = audio_data->softBufSize2;
	if(audio_data->buffer2IsEmpty) {
		while(!audio_data->bufferIsFull)
			usleep(10000);
		//memcpy(audio_data->buffer2, audio_data->buffer, audio_data->softBufSize);
		if((err = AudioConverterConvertBuffer(audio_data->converter, audio_data->softBufSize, audio_data->buffer, &converted, audio_data->buffer2)) != noErr) {
			printf("conversion error\n");
			AudioConverterDispose(audio_data->converter);
			exit(0);
		}
		audio_data->myBufferRest = audio_data->bufferRest*8/audio_data->channels/audio_data->bytesPerChannel;
		if(audio_data->myBufferRest < audio_data->buf_size) {
			audio_data->nextReadSize = audio_data->myBufferRest;
			audio_data->lastPlay = 1;
		}
		audio_data->buffer2IsEmpty = 0;
		audio_data->bufferIsFull = 0;
	}
	
	buffer = (float *)data_out->mBuffers [0].mData ;
	
	audio_data->nbytes = audio_data->nbytes + audio_data->nextReadSize*audio_data->channels*audio_data->bytesPerChannel/8;
	
	if(audio_data->channels == 1) {
		for(i=0;i<audio_data->nextReadSize/sizeof(float);i=i+2) {
			*(buffer+i) = *(audio_data->nextStart+i);
			*(buffer+i+1) = *(audio_data->nextStart+i);
		}
	}
	else 
		memcpy(buffer, audio_data->nextStart, audio_data->nextReadSize);
	audio_data->myBufferRest = audio_data->myBufferRest - audio_data->nextReadSize;
	if(!audio_data->lastPlay) audio_data->nextReadSize = audio_data->buf_size;
	audio_data->nextStart = audio_data->nextStart + audio_data->nextReadSize/sizeof(float);
	
	if(audio_data->nbytes == audio_data->dataBytes || audio_data->lastPlay) {
		memset (&(buffer[audio_data->nextReadSize/sizeof(float)]), 0, audio_data->buf_size - audio_data->nextReadSize) ;
		/* Tell the main application to terminate. */
		audio_data->done_playing = 1 ;
	}
	
	if(!audio_data->done_playing) {
		if(audio_data->myBufferRest == 0) {
			audio_data->buffer2IsEmpty = 1;
			audio_data->nextStart = audio_data->buffer2;
		}
		else if(audio_data->myBufferRest < audio_data->buf_size) {
			audio_data->nextReadSize = audio_data->myBufferRest;
			audio_data->lastPlay = 1;
		}
	}
	
	//printf("%d\n",audio_data->myBufferRest);
	if(!audio_data->quiet) {
		percent = (int)(100.0*audio_data->nbytes/(audio_data->dataBytes));
		
		if(percent != audio_data->percent) {
			printf("\r|");
			for(i=0;i<20;i++) {
				if(percent/5 > i)
					printf("=");
				else if(percent/5 == i)
					printf(">");
				else printf("-");
			}
			//printf("| %d%% %d/%d",percent,audio_data->nbytes,audio_data->dataBytes);
			printf("| %d%%",percent);
			fflush(stdout);
		}
		audio_data->percent = percent;
	}
	
	
	return noErr;
}
